'''
Permafrost modeling using ML, Revision 
S. Filhol, June 2021, revised September 2022

1. Load data
    - training dataset
    - maps
2. Define several BTS threshold target np.arange[-1.6, -3.4, -0.2]
3. Standardize/Scale the data if needed 
5. Split training dataset in train/test
6. Apply resampling (over-sampling) if needed
7. Fit ML models to training datasets
8. Plot model metrics
9. select few models
10. transform prediction data with scaler
11. predictNeu
12. plot results

Store data and move on statistical analysis of results


'''
#%%
# Import all libraries
import pdb

import pandas as pd
import xarray as xr
import geopandas as gpd
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import datetime as dt
plt.rcParams['figure.figsize'] = (15, 5)

# my personal plotting style
#plt.style.use('/home/simonfi/github/myDotFiles/matplotlib_style/style_IoT.mplstyle')

# import DEMs for the Retezat and Fagaras Mountain ranges
import rasterio

# import local modules
import logistic_regression as lr
import raster_utils as rt

# hide all pandas warnings
import warnings
warnings.filterwarnings('ignore')

from imblearn.over_sampling import RandomOverSampler
from sklearn.metrics import plot_confusion_matrix, plot_roc_curve, f1_score, roc_auc_score
from sklearn.preprocessing import RobustScaler
from sklearn.model_selection import train_test_split
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from sklearn import neighbors
from sklearn.datasets import make_moons, make_circles, make_classification
from sklearn.neural_network import MLPClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.metrics import RocCurveDisplay
from sklearn.inspection import permutation_importance
import shap
from sklearn.model_selection import StratifiedShuffleSplit



import os
#%%

usr = os.environ.get('USER')
work_dir = './'
output_path = 'outputs/'
if not os.path.isdir(work_dir + output_path):
    os.mkdir(work_dir + output_path)

#%%
#================================================================================
#           Populate the massif class object with all parameters
#================================================================================

retezat = rt.massif(name='retezat', fullname='Retezat', pwd=work_dir)
    
retezat.extent = [305200, 413830, 356200, 446830]
retezat.dem = './pm_inputs/retezat_dem.tif'
retezat.slope = './pm_inputs/retezat_slope.tif'
retezat.mask = './pm_inputs/retezat_mask.tif'
retezat.ndvi = './pm_inputs/retezat_NDVI.tif'
retezat.solar = './pm_inputs/retezat_solar.tif'
retezat.snow_prob_median = './pm_inputs/retezat_snow.tif'
retezat.curv = './pm_inputs/retezat_curv.tif'
retezat.aspect = './pm_inputs/retezat_aspect.tif'
retezat._update_raster_dict()

parang = rt.massif(name='parang', fullname='Parâng', pwd=work_dir)
parang.extent = [377560, 420300, 398560, 435300]
parang.dem = './pm_inputs/parang_dem.tif'
parang.slope = './pm_inputs/parang_slope.tif'
parang.mask = './pm_inputs/parang_mask.tif'
parang.ndvi = './pm_inputs/parang_NDVI.tif'
parang.solar = './pm_inputs/parang_solar.tif'
parang.snow_prob_median = './pm_inputs/parang_snow.tif'
parang.curv = './pm_inputs/parang_curv.tif'
parang.aspect = './pm_inputs/parang_aspect.tif'

parang._update_raster_dict()

fagaras = rt.massif(name='fagaras-iezer', fullname='Făgăraș-Iezer', pwd=work_dir)
fagaras.extent = [447300, 432600, 510100, 468400]
fagaras.dem = './pm_inputs/fagaras-iezer_dem.tif'
fagaras.slope = './pm_inputs/fagaras-iezer_slope.tif'
fagaras.mask = './pm_inputs/fagaras-iezer_mask.tif'
fagaras.ndvi = './pm_inputs/fagaras-iezer_NDVI.tif'
fagaras.solar = './pm_inputs/fagaras-iezer_solar.tif'
fagaras.snow_prob_median = './pm_inputs/fagaras-iezer_snow.tif'
fagaras.curv = './pm_inputs/fagaras-iezer_curv.tif'
fagaras.aspect = './pm_inputs/fagaras-iezer_aspect.tif'

fagaras._update_raster_dict()

massif_list = [retezat, parang, fagaras]


#%%
#================================================================================
#                   Loading BTS data
#================================================================================
bts = pd.read_csv('./pm_inputs/BTS _Carpathians_2021_epsg3844.csv')
var_input_list = ['dem', 
            'ndvi',
            'solar',
            'slope',
            'snow_prob_median',
            'curv']
for var in var_input_list:
    bts[var] = np.nan

#%%
# Sample datasets for each massif at location of the BTS measurements
for massif in massif_list:
    massif.get_rasterio_meta()
    for val in var_input_list:
        bts.loc[bts.Massif==massif.fullname] = rt.sample_raster(bts.loc[bts.Massif==massif.fullname], massif.raster_dict[val], col_name=val)

# groupby elevation band and BTS values
elev_band = np.arange(1850, 2400, 50)
bts['Elevation_band'] = pd.cut(bts.dem, bins=elev_band)
bts_band = np.arange(-13, 0, 1)
bts['BTS_band'] = pd.cut(bts.BTS, bins=bts_band)

thresholds = [-2, -2.5, -3]

bts['permafrost_type'] = None
bts['permafrost_type'].loc[bts.BTS > -2] = 'None'
bts['permafrost_type'].loc[(bts.BTS > -3) & (bts.BTS <= -2)] = 'Discontinuous'
bts['permafrost_type'].loc[bts.BTS <= -3] = 'Continuous'

# Classify for discontinous permafrost vs. none
for val in thresholds:
    bts['perm_IO_' + str(-val)] = (bts.BTS <= val) * 1

#bts.to_csv(massif.pwd + output_path + 'bts_measurements.csv')
bts.to_pickle(massif.pwd + output_path + 'bts_measurements.pkl')
#%%
# Scale the data
# need to loop through the threshold

def scale_df(df, cols=['dem', 'solar','slope'], transformer=None):
    # Function to scale down features
    if transformer is None:
        transformer = RobustScaler().fit(df[cols])
    dd = df.copy()
    dd[cols] = transformer.transform(dd[cols])
    dd.snow_prob_median = dd.snow_prob_median / 100
    #dd.snow_prob_std = dd.snow_prob_std / 10
    
    return dd, transformer

def inverse_scale_df(df, cols=['dem', 'solar', 'slope'], transformer=None):
    # function to inverse scaling
    dd = df.copy()
    dd[cols] = transformer.inverse_transform(df[cols])
    dd.snow_prob_median = dd.snow_prob_median * 100
    #dd.snow_prob_std = dd.snow_prob_std * 10
    return dd
    
#%%
#================================================================================
#---------- Temperature threshold sensitivity ------------------------
#================================================================================

# Initiate a general dictionnary for each threshold. Dict adapts to number of thresholds defined in
thres_dfs = {}
for thres in thresholds:
    thres_dfs.update({'perm_IO_{}'.format(np.abs(thres)):None})

input_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']
feature_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']

for thres in thres_dfs:
    df = bts[input_list + [thres] + ['X', 'Y']]
    df = df.dropna(axis=0)
    df = df.rename(columns={thres: 'target'})

    # scale the data
    df, scaler = scale_df(df)
    
    # Compute weights based on number of neighbors within a given radius
    knn = neighbors.NearestNeighbors(radius=100)  # radius in meters
    knn.fit(df[['X', 'Y']])
    nb = knn.radius_neighbors()[1]
    n = []
    for pt in nb:
        n.append(len(pt))
    n = np.array(n)
    #print(n)
    df['weights'] = (1/(n+1))/np.sum(1/(n+1))

    #pdb.set_trace()
    # split training data set for testing
    random_seed = 888
    df_train, df_test = lr.train_test_split(df, test_size=0.3, random_state=random_seed, stratify=df['target'])

    thres_dfs[thres] = {'df_obs': df, 'scaler': scaler, 'df_train': df_train, 'df_test': df_test}



models = {'Logistic Regression': LogisticRegression(solver='liblinear', max_iter=500),
          'Random Forest': RandomForestClassifier(max_depth=6, max_features=4, n_estimators=100),
          'Neural Net': MLPClassifier(max_iter=3000, activation='relu', solver='adam', hidden_layer_sizes=100),
          'Gradient Boost': GradientBoostingClassifier(),
          'AdaBoost': AdaBoostClassifier(learning_rate= 0.05, n_estimators= 500),
          'KNeighbor': KNeighborsClassifier(15, weights='distance'),
          'SVM': SVC(C=500, probability=True)
         }

for massif in massif_list:
    massif.to_df(var=input_list)

    # Need to handle NaN values.
    #   - Slope has one border of pixels at -9999
    #   - solar has 
    if 'slope' in input_list:
        massif.df['slope'].loc[massif.df.slope == -9999] = 0
    if 'solar' in input_list:
        massif.df['solar'].loc[massif.df.solar < 0] = massif.df.solar.median()

    dd_scaled = scale_df(massif.df[input_list], transformer=scaler)[0]
    massif.df = pd.concat([dd_scaled, massif.df[['X','Y']]], axis=1)
    dd_scaled = None        #free memory
    massif.models = models
           
    
#%%
# ==== Plotting ROCs

colors = [(64/255, 83/255, 211/255), 
          (221/255, 179/255, 16/255), 
          (181/255, 29/255, 20/255), 
          (0/255, 190/255, 255/255), 
          (251/255, 73/255, 176/255), 
          (0/255, 178/255, 93/255), 
          (202/255, 202/255, 202/255), 
          (102/255, 102/255, 102/255), 
          (2/255, 2/255, 2/255)]

fig, axs = plt.subplots(1,3)
for i, thres in enumerate(thres_dfs):
    for j, model in enumerate(models.keys()):
        print('... ... ', model)
        print('number of training samples = {}'.format(thres_dfs[thres]['df_train'].shape[0]))
        
        # fitting modle to data
        if model in ['KNeighbor', 'Neural Net']:
            models[model].fit(thres_dfs[thres]['df_train'][feature_list], thres_dfs[thres]['df_train'].target)
            trash = RocCurveDisplay.from_estimator(models[model], thres_dfs[thres]['df_test'][feature_list], thres_dfs[thres]['df_test'].target, ax=axs[i], color = colors[j])
        else:
            models[model].fit(thres_dfs[thres]['df_train'][feature_list], thres_dfs[thres]['df_train'].target, sample_weight=thres_dfs[thres]['df_train']['weights'])

            trash = RocCurveDisplay.from_estimator(models[model], thres_dfs[thres]['df_test'][feature_list], thres_dfs[thres]['df_test'].target, sample_weight=thres_dfs[thres]['df_test']['weights'], ax=axs[i],
                                                  color = colors[j])
    axs[i].plot((0,1),(0,1), linestyle='--', c='k', linewidth=2)
    axs[i].set_title(f'Threshold: {thresholds[i]}' + ' $^{o}C$')
    if i > 0:
        axs[i].set_ylabel(None)
        axs[i].set_yticklabels([])

#%%
# fitting models for each massif and each thresholds, storing outpur into a pandas dataframe

for massif in massif_list:
    print('---> Modeling massif: {}'.format(massif.name))
    for i, thres in enumerate(thres_dfs):
        print('... threshhold: {}'.format(thres))
        for j, model in enumerate(models.keys()):
            print('... ... ', model)
            print('number of training samples = {}'.format(thres_dfs[thres]['df_train'].shape[0]))

            # fitting modle to data
            if model in ['KNeighbor', 'Neural Net']:
                models[model].fit(thres_dfs[thres]['df_train'][feature_list], thres_dfs[thres]['df_train'].target)

            else:
                models[model].fit(thres_dfs[thres]['df_train'][feature_list], thres_dfs[thres]['df_train'].target, sample_weight=thres_dfs[thres]['df_train']['weights'])

            # Predict probability
            massif.df[thres + '_' + model] = models[model].predict_proba(massif.df[feature_list])[:, 1]
            
    # add mask as a column to the dataframe
    with rasterio.open(massif.mask) as src:
        arr = np.flip(src.read(1),0)
    massif.df['mask'] = arr.flatten()
    massif.df['mask'].loc[massif.df['mask']<0] = 0
    
    # inverse back matrix
    massif.df[input_list] = inverse_scale_df(massif.df[input_list], transformer=scaler)
        
    
        
#%%
# Store results in netcdf

# store input variables into a single netcdf file for each massif
for massif in massif_list:
    print('Storing {} to nc'.format(massif.name))
    proba_map = np.zeros([massif.height, massif.width, len(
        thresholds), len(list(models.keys()))])
    for i, thres in enumerate(thres_dfs):
        for k, model in enumerate(massif.models):
            Z = massif.df[thres + '_' + model].values.reshape(massif.height, massif.width)
            proba_map[:,:,i, k] = np.flip(Z, axis=0)
    
    offset = massif.raster_profile['transform'][0]/2
    x = np.arange(massif.raster_profile['transform'][2]+offset,
                  massif.raster_profile['transform'][2]+massif.width*massif.raster_profile['transform'][0]+offset,
                  massif.raster_profile['transform'][0])
    y = np.arange(massif.raster_profile['transform'][5]+massif.height*massif.raster_profile['transform'][4]+offset,
                  massif.raster_profile['transform'][5]+offset,
                  massif.raster_profile['transform'][0])
    
    ds_input = xr.Dataset(
        data_vars=dict(
            permafrost_probability=(["y", "x", "threshold","model"], proba_map),
            elevation=(["y", "x"], np.flip(massif.df['dem'].values.reshape(massif.height, massif.width), 0)),
            slope=(["y", "x"], np.flip(massif.df['slope'].values.reshape(massif.height, massif.width), 0)),
            mask=(["y", "x"], massif.df['mask'].values.reshape(massif.height, massif.width)),
            solar=(["y", "x"], np.flip(massif.df['solar'].values.reshape(massif.height, massif.width), 0)),
            ndvi=(["y", "x"], np.flip(massif.df['ndvi'].values.reshape(massif.height, massif.width), 0)),
            curv=(["y", "x"], np.flip(massif.df['curv'].values.reshape(massif.height, massif.width), 0)),
            snow_prob_median=(["y", "x"], np.flip(massif.df['snow_prob_median'].values.reshape(massif.height, massif.width), 0))
        ),
        coords={"x": x, "y": y, "threshold": thresholds, "model": list(models.keys())},
        attrs={'description': "Permafrost extent modeling in the Romanian Carpathians",
              'author': 'Simon Filhol',
              'creation_date': dt.datetime.now().strftime('%Y/%m/%d %H:%M:%S'),
              'massif': massif.name},
    )
            
    fout = massif.pwd + output_path + massif.name + '_all.nc'
    encode = {
        "permafrost_probability": {"compression": "gzip", "compression_opts": 9},
        "elevation": {"compression": "gzip", "compression_opts": 9},
        "slope": {"compression": "gzip", "compression_opts": 9},
        "mask": {"compression": "gzip", "compression_opts": 9},
        "solar": {"compression": "gzip", "compression_opts": 9},
        "ndvi": {"compression": "gzip", "compression_opts": 9},
        "curv": {"compression": "gzip", "compression_opts": 9},
        "snow_prob_median": {"compression": "gzip", "compression_opts": 9}
    }
    ds_input.elevation.attrs = {'units': 'm'}
    ds_input.slope.attrs = {'units': 'deg'}
    ds_input.solar.attrs = {'units': 'Wh/m2'}
    ds_input.ndvi.attrs = {'units': 'ratio'}
    ds_input.snow_prob_median.attrs = {'units': 'probability'}
    ds_input.permafrost_probability.attrs = {'units': 'probability'}
    ds_input.to_netcdf(fout, 'w', encoding=encode, engine='h5netcdf')
    
#%%
# === Learning curve

from sklearn.model_selection import learning_curve

training_size = np.arange(100,750,100)
cv = 15

fig, axs = plt.subplots(models.__len__(),3)#, sharex=True, sharey=True)
for i, thres in enumerate(thres_dfs):
    for j, model in enumerate(models.keys()):
        if model in ['KNeighbor', 'Neural Net']:
            train_sizes, train_scores, valid_scores = learning_curve(models[model], 
                                                                     thres_dfs[thres]['df_obs'][feature_list], 
                                                                     thres_dfs[thres]['df_obs'].target,
                                                                     train_sizes=training_size, 
                                                                     cv=cv)
        else:
            train_sizes, train_scores, valid_scores = learning_curve(models[model], 
                                                                     thres_dfs[thres]['df_obs'][feature_list], 
                                                                     thres_dfs[thres]['df_obs'].target,
                                                                     train_sizes=training_size, 
                                                                     cv=cv, 
                                                                     fit_params= {'sample_weight':thres_dfs[thres]['df_obs']['weights']})

        
        axs[j,i].fill_between(training_size, 
                              np.mean(train_scores, axis=1)+np.std(train_scores, axis=1),
                              np.mean(train_scores, axis=1)-np.std(train_scores, axis=1), color='g', alpha=.2)
        axs[j,i].plot(training_size, np.mean(train_scores, axis=1), 'o-', c='g', label='Training score')
        
        axs[j,i].fill_between(training_size, 
                              np.mean(valid_scores, axis=1)+np.std(valid_scores, axis=1),
                              np.mean(valid_scores, axis=1)-np.std(valid_scores, axis=1),
                              color='b', alpha=.2)
        axs[j,i].plot(training_size, np.mean(valid_scores, axis=1), 'o-', c='b', label='Cross-validation score')
        
for i, thres in enumerate(thres_dfs):
    for j, model in enumerate(models.keys()):
        if i==0:
            axs[j,i].set_yticks([0.5,0.6,0.7,0.8,0.9,1])
        else:
            axs[j,i].set_yticks([0.5,0.6,0.7,0.8,0.9,1])
            axs[j,i].set_yticklabels([])
       
        if j==models.__len__()-1:
            axs[j,i].set_xticks(np.arange(0,800,100))
            axs[j,i].set_xlabel('Training size')
        else:
            axs[j,i].set_xticks(training_size)
            axs[j,i].set_xticklabels([])
        if j==0:
            axs[j,i].set_title(f'Threshold: {thresholds[i]}' + ' $^{o}C$')
        if i==0:
            axs[j,i].set_ylabel(model+'\n' + 'Score')

            
plt.legend()        
plt.show()


#%%
#====================================================================================================
#                                  Feature importance
#====================================================================================================

# method adapted from https://inria.github.io/scikit-learn-mooc/python_scripts/dev_features_importance.html#feature-importance-by-permutation

models = {'Logistic Regression': LogisticRegression(solver='liblinear', max_iter=500),
          'Random Forest': RandomForestClassifier(max_depth=6, max_features=4, n_estimators=100),
          'Neural Net': MLPClassifier(max_iter=3000, activation='relu', solver='adam', hidden_layer_sizes=100),
          'Gradient Boost': GradientBoostingClassifier(),
          'AdaBoost': AdaBoostClassifier(learning_rate= 0.05, n_estimators= 500),
          'KNeighbor': KNeighborsClassifier(15, weights='distance'),
          'SVM': SVC(C=500, probability=True)
         }

input_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']
feature_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']
df = bts[input_list + ['X', 'Y', 'perm_IO_2']]
df = df.dropna(axis=0)
df = df.rename(columns={'perm_IO_2': 'target'})

# scale the data
df, scaler = scale_df(df)

# Compute weights based on number of neighbors within a given radius
knn = neighbors.NearestNeighbors(radius=100)  # radius in meters
knn.fit(df[['X', 'Y']])
nb = knn.radius_neighbors()[1]
n = []
for pt in nb:
    n.append(len(pt))
n = np.array(n)
#print(n)
df['weights'] = (1/(n+1))/np.sum(1/(n+1))
df = df.rename(columns={'dem':'Elevation', 'slope':'Slope', 'curv':'Curvature', 'snow_prob_median':'SPI', 'ndvi':'NDVI', 'solar':'PISR'})
feature_list = ['Elevation', 'Slope', 'PISR', 'NDVI', 'Curvature', 'SPI']

#pdb.set_trace()
# split training data set for testing
random_seed = 888
df_train, df_test = lr.train_test_split(df, test_size=0.3, random_state=random_seed, stratify=df['target'])

for j, model in enumerate(models.keys()):
    if model in ['KNeighbor', 'Neural Net']:
        models[model].fit(df_train[feature_list], df_train.target)
        print('---> {} score on testing data: {}'.format(model, models[model].score(df_test[feature_list], df_test.target)))
    else:
        models[model].fit(df_train[feature_list], df_train.target, sample_weight=df_train['weights'])
        print('---> {} score on testing data: {}'.format(model, models[model].score(df_test[feature_list], df_test.target, sample_weight=df_test['weights'])))
        

# plot the two score in one figure

fig, axs = plt.subplots(len(models.keys()), 2, sharex= True, figsize=(14, 20))

X_train = df_train[feature_list].copy().reset_index().drop(columns='index')
X_test = df_test[feature_list].copy().reset_index().drop(columns='index')

# adding a random feature as reference
rng = np.random.RandomState(0)
bin_var = pd.Series(rng.randint(0, 1, X_train.shape[0]), name='rnd_bin')
X_train = pd.concat((X_train, bin_var), axis=1)
bin_var = pd.Series(rng.randint(0, 1, X_test.shape[0]), name='rnd_bin')
X_test = pd.concat((X_test, bin_var), axis=1)
y_train = df_train.target.copy().reset_index().drop(columns='index')
y_test = df_test.target.copy().reset_index().drop(columns='index')

for j, model in enumerate(models.keys()):
    print(model)
    n_repeats=100

    if model in ['KNeighbor', 'Neural Net']:
        models[model].fit(X_train, y_train)
        r = permutation_importance(models[model], X_test, y_test, n_repeats=n_repeats, random_state=0, scoring='roc_auc')
    else:
        models[model].fit(X_train, y_train, sample_weight=df_train['weights'])
        r = permutation_importance(models[model], X_test, y_test, n_repeats=n_repeats, random_state=0, scoring='roc_auc')

    indices = r['importances_mean'].argsort()
    axs[j,0].barh(range(len(indices)),
             r['importances_mean'][indices],
             xerr=r['importances_std'][indices])

    axs[j,0].set_yticks(range(len(indices)))
    _ = axs[j,0].set_yticklabels(X_test.columns[indices])
    axs[j,0].set_title(model)
for j, model in enumerate(models.keys()):
    print(model)
    n_repeats=100

    if model in ['KNeighbor', 'Neural Net']:
        models[model].fit(X_train, y_train)
        r = permutation_importance(models[model], X_test, y_test, n_repeats=n_repeats, random_state=0, scoring='r2')
    else:
        models[model].fit(X_train, y_train, sample_weight=df_train['weights'])
        r = permutation_importance(models[model], X_test, y_test, n_repeats=n_repeats, random_state=0, scoring='r2')

    indices = r['importances_mean'].argsort()
    axs[j,1].barh(range(len(indices)),
             r['importances_mean'][indices],
             xerr=r['importances_std'][indices])

    axs[j,1].set_yticks(range(len(indices)))
    _ = axs[j,1].set_yticklabels(X_test.columns[indices])
    axs[j,1].set_title(model)

plt.tight_layout()
plt.show()
#X_train=None
#y_train=None
#X_test=None
#y_test=None

for j, model in enumerate(models.keys()):
    if j==models.__len__()-1:
        axs[j].set_xlabel('$R^{2} score$')

        
        
#------------------------------------------------------------------
#         Feature Importance SHAP analysis
#------------------------------------------------------------------
 
        
plt.figure()
for j, model in enumerate(models.keys()):
    print('... ... ', model)
    print('number of training samples = {}'.format(df_train.shape[0]))

    # fitting modle to data
    if model in ['KNeighbor', 'Neural Net']:
        mod = models[model].fit(df_train[feature_list], df_train.target)
    else:
        mod = models[model].fit(df_train[feature_list], df_train.target, sample_weight=df_train['weights'])


    explainer = shap.Explainer(mod.predict, df_test[feature_list])
    shap_values = explainer(df_test[feature_list])

    shap.plots.beeswarm(shap_values)
plt.show()

#%%
#====================================================================================================
#         ROC Curves and Comfusion matrices for -2 threshold
#====================================================================================================
models = {'Logistic Regression': LogisticRegression(solver='liblinear', max_iter=500),
          'Random Forest': RandomForestClassifier(max_depth=6, max_features=4, n_estimators=100),
          'Neural Net': MLPClassifier(max_iter=3000, activation='relu', solver='adam', hidden_layer_sizes=100),
          'Gradient Boost': GradientBoostingClassifier(),
          'AdaBoost': AdaBoostClassifier(learning_rate= 0.05, n_estimators= 500),
          'KNeighbor': KNeighborsClassifier(15, weights='distance'),
          'SVM': SVC(C=500, probability=True)
         }

colors = [(64/255, 83/255, 211/255), 
          (221/255, 179/255, 16/255), 
          (181/255, 29/255, 20/255), 
          (0/255, 190/255, 255/255), 
          (251/255, 73/255, 176/255), 
          (0/255, 178/255, 93/255), 
          (202/255, 202/255, 202/255), 
          (102/255, 102/255, 102/255), 
          (2/255, 2/255, 2/255)]

input_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']
feature_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']
df = bts[input_list + ['X', 'Y', 'perm_IO_2']]
df = df.dropna(axis=0)
df = df.rename(columns={'perm_IO_2': 'target'})

# scale the data
df, scaler = scale_df(df)

# Compute weights based on number of neighbors within a given radius
knn = neighbors.NearestNeighbors(radius=100)  # radius in meters
knn.fit(df[['X', 'Y']])
nb = knn.radius_neighbors()[1]
n = []
for pt in nb:
    n.append(len(pt))
n = np.array(n)
#print(n)
df['weights'] = (1/(n+1))/np.sum(1/(n+1))
#df = df.rename(columns={'dem':'Elevation', 'slope':'Slope', 'curv':'Curvature', 'snow_prob_median':'SPI', 'ndvi':'NDVI', 'solar':'PISR'})
#feature_list = ['Elevation', 'Slope', 'PISR', 'NDVI', 'Curvature', 'SPI']

# load Maxent data
maxent_roc = pd.read_csv('./permafrost_modeling/maxent_ROC_values.csv')

#pdb.set_trace()
# split training data set for testing
random_seed = 888
df_train, df_test = lr.train_test_split(df, test_size=0.3, random_state=random_seed, stratify=df['target'])

fig, axs = plt.subplots(1, 1, figsize=(10,10))
fi1, ax1 = plt.subplots(1,len(models.keys()), figsize=(30,8))
axs.plot(maxent_roc['False Positive rate'], maxent_roc['True Positive Rate'], label='Maxent (AUC=0.995)')
axs.plot((0,1),(0,1), linestyle='--', c='k', linewidth=2, label='Random guess (AUC=0.5)')
for j, model in enumerate(models.keys()):
    print('... ... ', model)
    print('number of training samples = {}'.format(df_train.shape[0]))

    # fitting modle to data
    if model in ['KNeighbor', 'Neural Net']:
        models[model].fit(df_train[feature_list], df_train.target)
        auc = roc_auc_score(df_test.target, models[model].predict_proba(df_test[feature_list])[:, 1])
        
        trash = RocCurveDisplay.from_estimator(models[model], df_test[feature_list], df_test.target, ax=axs, color = colors[j], label=f'{model} (AUC={np.round(auc,2)})')
        
    else:
        models[model].fit(df_train[feature_list], df_train.target, sample_weight=df_train['weights'])
        
        # Compute AUC
        auc = roc_auc_score(df_test.target, models[model].predict_proba(df_test[feature_list])[:, 1], sample_weight=df_test['weights'])
        
        trash = RocCurveDisplay.from_estimator(models[model], df_test[feature_list], df_test.target, sample_weight=df_test['weights'], ax=axs,
                                              color = colors[j], label=f'{model} (AUC={np.round(auc,2)})')
    plot_confusion_matrix(models[model], df_test[feature_list], df_test.target, ax=ax1[j])
    ax1[j].set_title(f'{model}')

axs.plot((0,1),(0,1), linestyle='--', c='k', linewidth=2, label='Random guess (AUC=0.5)')
axs.set_title(f'{df_train.shape[0]} Training Samples')

for j, model in enumerate(models.keys()):
    if j>0:
        ax1[j].yaxis.set_visible(False)

plt.tight_layout()
plt.show()



#%%
#====================================================================================================
#         Reduce amount of training data from 100 to max points gradually
#====================================================================================================


input_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']
feature_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']

bags = [0.3,0.5,0.7, 0.8, 0.95]
bags_dfs = {}
for i in bags:
    bags_dfs.update({f'bts_{i}':None})


for i, bag in enumerate(bags_dfs):
    df = bts[input_list + ['X', 'Y', 'perm_IO_2']]
    df = df.dropna(axis=0)
    df = df.rename(columns={'perm_IO_2': 'target'})

    # scale the data
    df, scaler = scale_df(df)
    
    # Compute weights based on number of neighbors within a given radius
    knn = neighbors.NearestNeighbors(radius=100)  # radius in meters
    knn.fit(df[['X', 'Y']])
    nb = knn.radius_neighbors()[1]
    n = []
    for pt in nb:
        n.append(len(pt))
    n = np.array(n)
    #print(n)
    df['weights'] = (1/(n+1))/np.sum(1/(n+1))

    #pdb.set_trace()
    # split training data set for testing
    random_seed = 888
    df_train, df_test = lr.train_test_split(df, test_size=bags[i], random_state=random_seed, stratify=df['target'])

    bags_dfs[bag] = {'df_obs': df, 'scaler': scaler, 'df_train': df_train, 'df_test': df_test}


models = {'Logistic Regression': LogisticRegression(solver='liblinear', max_iter=500),
          'Random Forest': RandomForestClassifier(max_depth=6, max_features=4, n_estimators=100),
          'Neural Net': MLPClassifier(max_iter=3000, activation='relu', solver='adam', hidden_layer_sizes=100),
          'Gradient Boost': GradientBoostingClassifier(),
          'AdaBoost': AdaBoostClassifier(learning_rate= 0.05, n_estimators= 500),
          'KNeighbor': KNeighborsClassifier(15, weights='distance'),
          'SVM': SVC(C=500, probability=True)
         }

for massif in massif_list:
    massif.to_df(var=input_list)

    # Need to handle NaN values.
    #   - Slope has one border of pixels at -9999
    #   - solar has 
    if 'slope' in input_list:
        massif.df['slope'].loc[massif.df.slope == -9999] = 0
    if 'solar' in input_list:
        massif.df['solar'].loc[massif.df.solar < 0] = massif.df.solar.median()

    dd_scaled = scale_df(massif.df[input_list], transformer=scaler)[0]
    massif.df = pd.concat([dd_scaled, massif.df[['X','Y']]], axis=1)
    dd_scaled = None        #free memory
    massif.models = models
           
    
#%%
#========================================================================
#         ROC curves for different training sample size
#========================================================================

colors = [(64/255, 83/255, 211/255), 
          (221/255, 179/255, 16/255), 
          (181/255, 29/255, 20/255), 
          (0/255, 190/255, 255/255), 
          (251/255, 73/255, 176/255), 
          (0/255, 178/255, 93/255), 
          (202/255, 202/255, 202/255), 
          (102/255, 102/255, 102/255), 
          (2/255, 2/255, 2/255)]


# Plot per sample bag size. 
fig, axs = plt.subplots(1, 5, figsize=(30,10))
for i, bag in enumerate(bags_dfs):
    models = {'Logistic Regression': LogisticRegression(solver='liblinear', max_iter=500),
          'Random Forest': RandomForestClassifier(max_depth=6, max_features=4, n_estimators=100),
          'Neural Net': MLPClassifier(max_iter=3000, activation='relu', solver='adam', hidden_layer_sizes=100),
          'Gradient Boost': GradientBoostingClassifier(),
          'AdaBoost': AdaBoostClassifier(learning_rate= 0.05, n_estimators= 500),
          'KNeighbor': KNeighborsClassifier(15, weights='distance'),
          'SVM': SVC(C=500, probability=True)
         }
    print(f'bag {bag}: training samples: {bags_dfs[bag]["df_train"].shape[0]}')

    for j, model in enumerate(models.keys()):
        print('... ... ', model)
        print('number of training samples = {}'.format(bags_dfs[bag]['df_train'].shape[0]))
        
        
        # fitting modle to data
        if model in ['KNeighbor', 'Neural Net']:
            models[model].fit(bags_dfs[bag]['df_train'][feature_list], bags_dfs[bag]['df_train'].target)
            auc = roc_auc_score(bags_dfs[bag]['df_test'].target, models[model].predict_proba(bags_dfs[bag]['df_test'][feature_list])[:, 1])
            trash = RocCurveDisplay.from_estimator(models[model], 
                                                   bags_dfs[bag]['df_test'][feature_list], 
                                                   bags_dfs[bag]['df_test'].target, 
                                                   ax=axs[i], color = colors[j], label=f'{model} (AUC={np.round(auc,2)})')
        else:
            models[model].fit(bags_dfs[bag]['df_train'][feature_list], bags_dfs[bag]['df_train'].target, sample_weight=bags_dfs[bag]['df_train']['weights'])
            auc = roc_auc_score(bags_dfs[bag]['df_test'].target, models[model].predict_proba(bags_dfs[bag]['df_test'][feature_list])[:, 1], sample_weight=bags_dfs[bag]['df_test']['weights'])

            trash = RocCurveDisplay.from_estimator(models[model], 
                                                   bags_dfs[bag]['df_test'][feature_list], 
                                                   bags_dfs[bag]['df_test'].target, 
                                                   sample_weight=bags_dfs[bag]['df_test']['weights'], 
                                                   ax=axs[i], color = colors[j], label=f'{model} (AUC={np.round(auc,2)})')
        
        # Predict probability
        #for massif in massif_list[0:1]:
        #    print(f'--> prediction for {massif.name}')
        #    massif.df[bag + '_' + model] = models[model].predict_proba(massif.df[feature_list])[:, 1]
            
    # add mask as a column to the dataframe
    #with rasterio.open(massif.mask) as src:
    #    arr = np.flip(src.read(1),0)
    #massif.df['mask'] = arr.flatten()
    #massif.df['mask'].loc[massif.df['mask']<0] = 0
    
    #
    axs[i].plot((0,1),(0,1), linestyle='--', c='k', linewidth=2)
    axs[i].set_title(f'{bags_dfs[bag]["df_train"].shape[0]} Training Samples')
    if i > 0:
        axs[i].set_ylabel(None)
        axs[i].set_yticklabels([])
plt.tight_layout()
plt.subplots_adjust(hspace=0, wspace=0.01)
plt.show()

# Store results in netcdf
# store input variables into a single netcdf file for each massif
save_to_file=True
for massif in massif_list[0:1]:
    print('Storing {} to nc'.format(massif.name))
    proba_map = np.zeros([massif.height, massif.width, len(
        bags_dfs), len(list(models.keys()))])
    for i, bag in enumerate(bags_dfs):
        for k, model in enumerate(massif.models):
            Z = massif.df[bag + '_' + model].values.reshape(massif.height, massif.width)
            proba_map[:,:,i, k] = np.flip(Z, axis=0)
    
    offset = massif.raster_profile['transform'][0]/2
    x = np.arange(massif.raster_profile['transform'][2]+offset,
                  massif.raster_profile['transform'][2]+massif.width*massif.raster_profile['transform'][0]+offset,
                  massif.raster_profile['transform'][0])
    y = np.arange(massif.raster_profile['transform'][5]+massif.height*massif.raster_profile['transform'][4]+offset,
                  massif.raster_profile['transform'][5]+offset,
                  massif.raster_profile['transform'][0])
    
    ds_input = xr.Dataset(
        data_vars=dict(
            permafrost_probability=(["y", "x", "bag","model"], proba_map),
            elevation=(["y", "x"], np.flip(massif.df['dem'].values.reshape(massif.height, massif.width), 0)),
            slope=(["y", "x"], np.flip(massif.df['slope'].values.reshape(massif.height, massif.width), 0)),
            mask=(["y", "x"], massif.df['mask'].values.reshape(massif.height, massif.width)),
            solar=(["y", "x"], np.flip(massif.df['solar'].values.reshape(massif.height, massif.width), 0)),
            ndvi=(["y", "x"], np.flip(massif.df['ndvi'].values.reshape(massif.height, massif.width), 0)),
            curv=(["y", "x"], np.flip(massif.df['curv'].values.reshape(massif.height, massif.width), 0)),
            snow_prob_median=(["y", "x"], np.flip(massif.df['snow_prob_median'].values.reshape(massif.height, massif.width), 0))
        ),
        coords={"x": x, "y": y, "bag":  list(bags_dfs.keys()), "model": list(models.keys())},
        attrs={'description': "Permafrost extent modeling in the Romanian Carpathians",
              'author': 'Simon Filhol',
              'creation_date': dt.datetime.now().strftime('%Y/%m/%d %H:%M:%S'),
              'massif': massif.name},
    )
            
    fout = massif.pwd + output_path + massif.name + '_training_sample_size.nc'
    encode = {
        "permafrost_probability": {"compression": "gzip", "compression_opts": 9},
        "elevation": {"compression": "gzip", "compression_opts": 9},
        "slope": {"compression": "gzip", "compression_opts": 9},
        "mask": {"compression": "gzip", "compression_opts": 9},
        "solar": {"compression": "gzip", "compression_opts": 9},
        "ndvi": {"compression": "gzip", "compression_opts": 9},
        "curv": {"compression": "gzip", "compression_opts": 9},
        "snow_prob_median": {"compression": "gzip", "compression_opts": 9}
    }
    ds_input.elevation.attrs = {'units': 'm'}
    ds_input.slope.attrs = {'units': 'deg'}
    ds_input.solar.attrs = {'units': 'Wh/m2'}
    ds_input.ndvi.attrs = {'units': 'ratio'}
    ds_input.snow_prob_median.attrs = {'units': 'probability'}
    ds_input.permafrost_probability.attrs = {'units': 'probability'}
    if save_to_file:
        ds_input.to_netcdf(fout, 'w', encoding=encode, engine='h5netcdf')
   
    
# ['Logistic Regression', 'Random Forest', 'Neural Net', 'Gradient Boost', 'AdaBoost', 'KNeighbor', 'SVM']
fig, axs = plt.subplots(5,1, sharex=True, sharey=True, figsize=(10,20))
for i, b in enumerate(ds_input.bag.values):
    (ds_input.permafrost_probability.sel(bag=b, model='AdaBoost')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i], cmap=plt.cm.RdBu, vmin=0.4, vmax=0.6)
    axs[i].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
plt.tight_layout()
plt.show()

fig, axs = plt.subplots(5,1, sharex=True, sharey=True, figsize=(10,20))
for i, b in enumerate(ds_input.bag.values):
    (ds_input.permafrost_probability.sel(bag=b, model='Gradient Boost')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i], cmap=plt.cm.RdBu, vmin=0, vmax=1)
    axs[i].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
plt.tight_layout()
plt.show()

fig, axs = plt.subplots(5,1, sharex=True, sharey=True, figsize=(10,20))
for i, b in enumerate(ds_input.bag.values):
    (ds_input.permafrost_probability.sel(bag=b, model='Logistic Regression')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i], cmap=plt.cm.RdBu, vmin=0.4, vmax=0.6)
    axs[i].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
plt.tight_layout()
plt.show()

fig, axs = plt.subplots(5,1, sharex=True, sharey=True, figsize=(10,20))
for i, b in enumerate(ds_input.bag.values):
    (ds_input.permafrost_probability.sel(bag=b, model='Random Forest')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i], cmap=plt.cm.RdBu, vmin=0, vmax=1)
    axs[i].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
plt.tight_layout()
plt.show()

fig, axs = plt.subplots(5,1, sharex=True, sharey=True, figsize=(10,20))
for i, b in enumerate(ds_input.bag.values):
    (ds_input.permafrost_probability.sel(bag=b, model='Neural Net')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i], cmap=plt.cm.RdBu, vmin=0, vmax=1)
    axs[i].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
plt.tight_layout()
plt.show()

fig, axs = plt.subplots(5,1, sharex=True, sharey=True, figsize=(10,20))
for i, b in enumerate(ds_input.bag.values):
    (ds_input.permafrost_probability.sel(bag=b, model='KNeighbor')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i], cmap=plt.cm.RdBu, vmin=0, vmax=1)
    axs[i].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
plt.tight_layout()
plt.show()

fig, axs = plt.subplots(5,1, sharex=True, sharey=True, figsize=(10,20))
for i, b in enumerate(ds_input.bag.values):
    (ds_input.permafrost_probability.sel(bag=b, model='SVM')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i], cmap=plt.cm.RdBu, vmin=0, vmax=1)
    axs[i].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
plt.tight_layout()
plt.show()
    
fig, axs = plt.subplots(5,2, sharex=True, sharey=True, figsize=(40,25))
for i, b in enumerate(ds_input.bag.values):
    (ds_input.permafrost_probability.sel(bag=b).median(dim='model')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i,0], cmap=plt.cm.RdBu, vmin=0.3, vmax=0.7)
    (ds_input.permafrost_probability.sel(bag=b).std(dim='model')*ds_input.mask).sel(x=slice(328000,340000), y=slice(430000,435000)).plot(ax=axs[i,1], cmap=plt.cm.viridis, vmin=0, vmax=0.3)
    axs[i,0].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
    axs[i,1].set_title(f'{bags_dfs[b]["df_train"].shape[0]} Training sample ')
plt.tight_layout()
plt.show()
    
    
    
    
    
# Plot per model (visualize loss of model quality for each model
fig, axs = plt.subplots(1, len(models), figsize=(30,10))
for i, model in enumerate(models.keys()):
    for j, bag in enumerate(bags_dfs):
        
        print('... ... ', model)
        print('number of training samples = {}'.format(bags_dfs[bag]['df_train'].shape[0]))
        
        # fitting modle to data
        if model in ['KNeighbor', 'Neural Net']:
            models[model].fit(bags_dfs[bag]['df_train'][feature_list], bags_dfs[bag]['df_train'].target)
            trash = RocCurveDisplay.from_estimator(models[model], bags_dfs[bag]['df_test'][feature_list], bags_dfs[bag]['df_test'].target, ax=axs[i], color = colors[j], label=f'{bags_dfs[bag]["df_train"].shape[0]} Training Samples')
        else:
            models[model].fit(bags_dfs[bag]['df_train'][feature_list], bags_dfs[bag]['df_train'].target, sample_weight=bags_dfs[bag]['df_train']['weights'])

            trash = RocCurveDisplay.from_estimator(models[model], bags_dfs[bag]['df_test'][feature_list], bags_dfs[bag]['df_test'].target, sample_weight=bags_dfs[bag]['df_test']['weights'], ax=axs[i],
                                                  color = colors[j], label=f'{bags_dfs[bag]["df_train"].shape[0]} Training Samples')
    axs[i].plot((0,1),(0,1), linestyle='--', c='k', linewidth=2)
    axs[i].set_title(model)
    if i > 0:
        axs[i].set_ylabel(None)
        axs[i].set_yticklabels([])
plt.tight_layout()
plt.subplots_adjust(hspace=0, wspace=0.01)
plt.show()  
        

# Removing weigths to fitting. Plotting per sample bag size.
fig, axs = plt.subplots(1,5, figsize=(30,10))
for i, bag in enumerate(bags_dfs):

    for j, model in enumerate(models.keys()):
        print('... ... ', model)
        print('number of training samples = {}'.format(bags_dfs[bag]['df_train'].shape[0]))
        
        # fitting modle to data
        
        models[model].fit(bags_dfs[bag]['df_train'][feature_list], bags_dfs[bag]['df_train'].target)
        trash = RocCurveDisplay.from_estimator(models[model], bags_dfs[bag]['df_test'][feature_list], bags_dfs[bag]['df_test'].target, ax=axs[i], color = colors[j])

    axs[i].plot((0,1),(0,1), linestyle='--', c='k', linewidth=2)
    axs[i].set_title(f'{bags_dfs[bag]["df_train"].shape[0]} Training Samples')
    if i > 0:
        axs[i].set_ylabel(None)
        axs[i].set_yticklabels([])    
plt.tight_layout()
plt.subplots_adjust(hspace=0, wspace=0.01)
plt.show()


# Removing weigths to fitting. Plotting per model
fig, axs = plt.subplots(1,len(models), figsize=(30,10))

for i, model in enumerate(models.keys()):
    for j, bag in enumerate(bags_dfs):

        
        print('... ... ', model)
        print('number of training samples = {}'.format(bags_dfs[bag]['df_train'].shape[0]))
        
        # fitting modle to data
        
        models[model].fit(bags_dfs[bag]['df_train'][feature_list], bags_dfs[bag]['df_train'].target)

        trash = RocCurveDisplay.from_estimator(models[model], bags_dfs[bag]['df_test'][feature_list], bags_dfs[bag]['df_test'].target, ax=axs[i],
                                              color = colors[j], label=f'{bags_dfs[bag]["df_train"].shape[0]} Training Samples')
    axs[i].plot((0,1),(0,1), linestyle='--', c='k', linewidth=2)
    axs[i].set_title(model)
    if i > 0:
        axs[i].set_ylabel(None)
        axs[i].set_yticklabels([])
plt.tight_layout()
plt.subplots_adjust(hspace=0, wspace=0.01)
plt.show()


#%%
#====================================================================================================
#         ShuffleSplit 100 times of training data for 30% testing data 
#====================================================================================================   

input_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']
feature_list = ['dem', 'slope', 'solar', 'ndvi', 'curv', 'snow_prob_median']

n_shuffle = 100
ss = StratifiedShuffleSplit(n_splits=n_shuffle, random_state=2, test_size=0.3)


df = bts[input_list + ['X', 'Y', 'perm_IO_2']]
df = df.dropna(axis=0)
df = df.rename(columns={'perm_IO_2': 'target'})

# scale the data
df, scaler = scale_df(df)

# Compute weights based on number of neighbors within a given radius
knn = neighbors.NearestNeighbors(radius=100)  # radius in meters
knn.fit(df[['X', 'Y']])
nb = knn.radius_neighbors()[1]
n = []
for pt in nb:
    n.append(len(pt))
n = np.array(n)
#print(n)
df['weights'] = (1/(n+1))/np.sum(1/(n+1))

#pdb.set_trace()
# split training data set for testing
random_seed = 888
#df_train, df_test = lr.train_test_split(df, test_size=bags[i], random_state=random_seed, stratify=df['target'])


shufs = range(0,n_shuffle)
shufs_dfs = {}
for i in shufs:
    shufs_dfs.update({f'run_{i}':None})
    
i=0    
for train_index, test_index in ss.split(df[feature_list], y=df['target']):
    shufs_dfs[f'run_{i}'] = {'df_obs': df, 'scaler': scaler, 'df_train': df.iloc[train_index], 'df_test': df.iloc[test_index]}
    i += 1
        
models = {'Logistic Regression': LogisticRegression(solver='liblinear', max_iter=500),
          'Random Forest': RandomForestClassifier(max_depth=6, max_features=4, n_estimators=100),
          'Neural Net': MLPClassifier(max_iter=3000, activation='relu', solver='adam', hidden_layer_sizes=100),
          'Gradient Boost': GradientBoostingClassifier(),
          'AdaBoost': AdaBoostClassifier(learning_rate= 0.05, n_estimators= 500),
          'KNeighbor': KNeighborsClassifier(15, weights='distance'),
          'SVM': SVC(C=500, probability=True)
         }

for massif in massif_list:
    massif.to_df(var=input_list)

    # Need to handle NaN values.
    #   - Slope has one border of pixels at -9999
    #   - solar has 
    if 'slope' in input_list:
        massif.df['slope'].loc[massif.df.slope == -9999] = 0
    if 'solar' in input_list:
        massif.df['solar'].loc[massif.df.solar < 0] = massif.df.solar.median()

    dd_scaled = scale_df(massif.df[input_list], transformer=scaler)[0]
    massif.df = pd.concat([dd_scaled, massif.df[['X','Y']]], axis=1)
    dd_scaled = None        #free memory
    massif.models = models

from sklearn.metrics import RocCurveDisplay

colors = [(64/255, 83/255, 211/255), 
          (221/255, 179/255, 16/255), 
          (181/255, 29/255, 20/255), 
          (0/255, 190/255, 255/255), 
          (251/255, 73/255, 176/255), 
          (0/255, 178/255, 93/255), 
          (202/255, 202/255, 202/255), 
          (102/255, 102/255, 102/255), 
          (2/255, 2/255, 2/255)]

   
fig, axs = plt.subplots(1,len(models))

aucs = pd.DataFrame()
for i, model in enumerate(models.keys()):
    for j, shuf in enumerate(shufs_dfs):
        
        print('... ... ', model)
        print('number of training samples = {}'.format(shufs_dfs[shuf]['df_train'].shape[0]))
        
        # fitting model to data
        if model in ['KNeighbor', 'Neural Net']:
            models[model].fit(shufs_dfs[shuf]['df_train'][feature_list], shufs_dfs[shuf]['df_train'].target)
            trash = RocCurveDisplay.from_estimator(models[model], shufs_dfs[shuf]['df_test'][feature_list], shufs_dfs[shuf]['df_test'].target, ax=axs[i], color = 'k', alpha=0.1)
        else:
            models[model].fit(shufs_dfs[shuf]['df_train'][feature_list], shufs_dfs[shuf]['df_train'].target, sample_weight=shufs_dfs[shuf]['df_train']['weights'])

            trash = RocCurveDisplay.from_estimator(models[model], shufs_dfs[shuf]['df_test'][feature_list], shufs_dfs[shuf]['df_test'].target, sample_weight=shufs_dfs[shuf]['df_test']['weights'], ax=axs[i],
                                                  color = 'k', alpha=0.1)
        aucs = aucs.append({'model':model, 'auc':trash.roc_auc, 'shuffle':shuf},ignore_index=True)

    axs[i].plot((0,1),(0,1), linestyle='--', c='k', linewidth=2)
    axs[i].set_title(model)

auc_means = aucs.groupby('model').auc.mean()
auc_stds = aucs.groupby('model').auc.std()
    
for i, model in enumerate(models.keys()):
    axs[i].text(0.6,0.005,f'AUC = {auc_means[model].round(2)}$\pm${auc_stds[model].round(2)}',
                ha="center", 
                va="center",
                bbox=dict(boxstyle="round", pad=0.5, fc='w'))
    axs[i].get_legend().remove()
    axs[i].set_xlabel('False Positive Rate')
    if i > 0:
        axs[i].set_ylabel(None)
        axs[i].set_yticklabels([])
        
        
        







